"""Module for dealing with student groups."""
from urllib.parse import urlencode
from xml.etree import ElementTree

import requests

import ocflib.account.search as search
from ocflib.ucb.directory import name_by_calnet_uid


_API = {
    'BASE': 'https://studentgroupservice.sait-west.berkeley.edu/service.asmx',
    'SERVICE': {
        'ORGS': 'CalLinkOrganizations',
        'SIGNATORIES_BY_OID': 'CalLinkGroupSignatories',
        'SIGNAT_ACTIVE': 'SignatoriesActiveStudentGroups',
        'SIGNAT_ALL': 'SignatoriesStudentGroups'
    }
}


def list_groups(name='', oid='', status='', type='', category=''):
    """Return groups by a general CalLink search.

    >>> list_groups(name="facility")
    {
        46187: {
            'accounts': ['decal', 'linux', 'ggroup', 'group'],
            'email': 'devnull@ocf.berkeley.edu',
            'inactive': False,
            'name': 'Open Computing Facility',
            'primary_contact': {
                'email': 'ckuehl@berkeley.edu',
                'name': 'Chris Kuehl',
            },
            'short_name': 'OCF',
            'type': 'ASUC Government Program',
            'website': 'https://www.ocf.berkeley.edu/',
        },
    }
    """
    def parser(root):
        def parse(group):
            oid = int(group.findtext('OrganizationId'))
            group_type = group.findtext('TypeName')
            return oid, {
                'name': group.findtext('Name'),
                'accounts':
                    [] if oid == 0 else search.users_by_callink_oid(oid),
                'email': group.findtext('Email'),
                'website': group.findtext('ExternalWebsite'),
                'short_name': group.findtext('ShortName'),
                'primary_contact': {
                    'name': group.findtext('PrimaryContactName'),
                    'email': group.findtext('PrimaryContactCampusEmail'),
                },
                'type': group_type,
                'inactive': any(s in group_type.upper() for s in ['FROZEN GROUP', 'NEW ORG PENDING'])
            }

        xml_groups = root.findall('Items/Organization')
        return {oid: name for oid, name in map(parse, xml_groups)}

    return _get_osl({
        'name': name,
        'organizationId': str(oid),
        'status': status,
        'type': type,
        'category': category,
    }, _API['SERVICE']['ORGS'], parser)


def group_by_oid(oid):
    """Return the group with the given OID.

    >>> group_by_oid(46187)
    {
        'accounts': ['decal', 'linux', 'ggroup', 'group'],
        'email': 'devnull@ocf.berkeley.edu',
        'inactive': False,
        'name': 'Open Computing Facility',
        'primary_contact': {
            'email': 'ckuehl@berkeley.edu',
            'name': 'Chris Kuehl',
        },
        'short_name': 'OCF',
        'type': 'ASUC Government Program',
        'website': 'https://www.ocf.berkeley.edu/',
    }
    """
    result = list_groups(oid=oid)
    if not result:
        return None
    else:
        return result[oid]


# TODO: add option to not resolve accounts for speed
def signatories_for_group(oid):
    """Return signatories for a group, including name and OCF account.

    >>> signatories_for_group(46187)
    {646431: {'accounts': ['sanjayk'], 'name': 'Sanjay Krishnan'},
     872544: {'accounts': ['daradib'], 'name': 'Mr. Dara Adib'},
     1029873: {'accounts': ['kpengboy'], 'name': 'KEVIN YANG PENG'},
     1031366: {'accounts': ['mattmcal'], 'name': 'Matthew James McAllister'},
     1031553: {'accounts': ['willh'], 'name': 'WILLIAM HO'},
     1032668: {'accounts': ['nickimp'], 'name': 'NICHOLAS DANIEL IMPICCICHE'},
     1034192: {'accounts': ['ckuehl'], 'name': 'CHRISTOPHER B KUEHL'}}
    """
    def parser(root):
        def parse(student):
            uid = int(student.findtext('Username'))
            name = name_by_calnet_uid(uid)
            users = search.users_by_calnet_uid(uid)
            return uid, {'name': name, 'accounts': users}

        xml_members = root.findall('Items/Membership')
        return {uid: details for uid, details in map(parse, xml_members)}

    return _get_osl({'organizationId': oid},
                    _API['SERVICE']['SIGNATORIES_BY_OID'], parser)


# TODO: add option to not resolve accounts for speed
def groups_by_student_signat(uid, service=_API['SERVICE']['SIGNAT_ACTIVE']):
    """Return active groups a student is a signatory for.

    >>> groups_by_student_signat(1034192)
    {46187: {'name': 'Open Computing Facility', accounts: ['decal', 'linux']}}
    """
    def parser(root):
        def parse(group):
            oid = int(group.findtext('groupId'))
            return oid, {
                'name': group.findtext('groupName'),
                'accounts':
                    [] if oid == 0 else search.users_by_callink_oid(oid)}

        xml_groups = root.findall('StudentGroupData/StudentGroupDatum')
        return {oid: name for oid, name in map(parse, xml_groups)}
    return _get_osl({'UID': uid}, service, parser)


def groups_by_student_signat_all(uid):
    """Return all (active and inactive) groups a student is a signatory for."""
    return groups_by_student_signat(uid, service=_API['SERVICE']['SIGNAT_ALL'])


def _get_osl(query, service, parser):
    """Query web service for student group information in XML format.

    You should probably use one of the nicer methods instead."""

    url = '{}/{}?{}'.format(_API['BASE'], service, urlencode(query))

    r = requests.get(url, timeout=20)
    return _parse_osl(ElementTree.fromstring(r.text), parser)


def _parse_osl(root, parser):
    """Assemble Python dictionaries of groups from XML document"""
    if root.findtext('Succeeded') == 'false':
        error_reason = root.findtext('Reason', default='unknown reason')
        raise Exception('Lookup failed: ' + error_reason)

    return parser(root)



from collections import namedtuple
from datetime import date
from datetime import datetime
from datetime import timedelta
from hashlib import md5
from urllib.parse import urlencode

import yaml

from ocflib.account.search import user_attrs
from ocflib.account.utils import is_in_group
from ocflib.misc.mail import email_for_user


STAFF_HOURS_FILE = '/etc/ocf/staff_hours.yaml'

Hour = namedtuple('Hour', ['day', 'time', 'staff', 'cancelled'])


class Staffer(namedtuple('Staffer', ['user_name', 'real_name', 'position'])):

    def gravatar(self, size=100):
        email = email_for_user(self.user_name, check_exists=False)
        return 'https://www.gravatar.com/avatar/{hash}?{params}'.format(
            hash=md5(email.lower().encode('utf-8')).hexdigest(),
            params=urlencode({'d': 'mm', 's': size}),
        )


def _load_staff_hours():
    """Load staff hours, from the /etc/ocf folder."""
    with open(STAFF_HOURS_FILE, 'r') as f:
        return yaml.safe_load(f)


def get_staff_hours():
    staff_hours = _load_staff_hours()

    def position(uid):
        staff_position_dict = {entry['username']: entry['position'] for entry in staff_hours['staff-positions']}
        if uid in staff_position_dict:
            return staff_position_dict[uid]
        elif is_in_group(uid, 'ocfroot'):
            return 'Technical Manager'
        else:
            return 'Staff Member'

    staff_hour_list = []
    # If we try iterating through the keys in staff-hours, it's non-deterministic ordering.
    # Because of the schema, we are guaranteed that these days are always here.
    for day in ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']:
        if not staff_hours['staff-hours'][day]:
            continue

        for staff_hour in staff_hours['staff-hours'][day]:
            staff_hour_list.append(
                Hour(
                    day=day,
                    time=_parse_hour(staff_hour['time']),
                    staff=[
                        Staffer(
                            user_name=attrs['uid'][0],
                            real_name=_remove_middle_names(attrs['cn'][0]),
                            position=position(attrs['uid'][0]),
                        ) for attrs in map(user_attrs, staff_hour['staff'])
                    ],
                    cancelled=staff_hour.get('cancelled', False)
                )
            )

    return staff_hour_list


def _parse_hour(hour):
    """
    Converts a 2-element list of hours like ['11:00', '13:00'] to a string
    in 12-hour time, like '11:00AM - 1:00PM'.
    Needed for backwards compatibility with the old staff hours file.
    """
    return '{} - {}'.format(datetime.strptime(hour[0], '%H:%M').strftime('%-I:%M%p'),
                            datetime.strptime(hour[1], '%H:%M').strftime('%-I:%M%p'))


def _remove_middle_names(name):
    names = name.split(' ')
    return names[0] + ' ' + names[-1]


def get_staff_hours_soonest_first():
    today = date.today()
    days = [(today + timedelta(days=i)).strftime('%A') for i in range(7)]
    return sorted(get_staff_hours(), key=lambda hour: days.index(hour.day))



"""Email handling and sending"""
import email.mime.multipart
import email.mime.text
import inspect
import socket
import subprocess
from email.utils import parseaddr

from jinja2 import Environment
from jinja2 import PackageLoader

import ocflib.misc.validators as validators

SENDMAIL_PATH = '/usr/sbin/sendmail'
MAIL_ROOT = 'root@ocf.berkeley.edu'

MAIL_ROOT = 'root@ocf.berkeley.edu'
MAIL_FROM = 'Open Computing Facility <help@ocf.berkeley.edu>'
MAIL_SIGNATURE = """Thanks for flying OCF,
The friendly staff of 171 MLK Student Union

=========
The Open Computing Facility is an all-volunteer, student-run service
group providing free printing, web hosting, disk space, and Unix shell
accounts.

We love free & open-source software. Sound like you? Get involved!
    https://ocf.io/staff

OCF volunteers hold weekly staff hours to provide support:
    https://ocf.io/staff-hours

Need help connecting to the OCF?
    https://ocf.io/ssh

Need to reset your account password?
    https://ocf.io/password"""

jinja_mail_env = Environment(loader=PackageLoader('ocflib', ''))
jinja_mail_env.globals = {
    'mail_signature': MAIL_SIGNATURE,
}


def email_for_user(username, check_exists=True):
    """Return email for a user.

    Currently, just appends @ocf.berkeley.edu, but could eventually do
    something more complicated.
    """
    if check_exists:
        from ocflib.account.search import user_exists
        if not user_exists(username):
            raise ValueError('Account "{}" does not exist.'.format(username))

    return '{}@ocf.berkeley.edu'.format(username)


def send_mail_user(user, subject, body, html_body=None, sender=MAIL_FROM):
    """Send a plan-text mail message to a user."""
    send_mail(email_for_user(user), subject, body, html_body=html_body, sender=sender)


def send_mail(to, subject, body, *, html_body=None, cc=None, sender=MAIL_FROM):
    """Send a plain-text mail message.

    `body` should be a string with newlines, wrapped at about 80 characters."""

    if not validators.valid_email(parseaddr(sender)[1]):
        raise ValueError('Invalid sender address.')

    if not validators.valid_email(parseaddr(to)[1]):
        raise ValueError('Invalid recipient address.')

    # create a container with content type multipart/alternative
    msg = email.mime.multipart.MIMEMultipart('alternative')

    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = to
    msg['Cc'] = cc

    # the non-html message, for if the client doesn't support html
    plain = email.mime.text.MIMEText(body, 'plain')
    msg.attach(plain)
    if html_body:
        # the "pretty" message, for when it does support html
        html = email.mime.text.MIMEText(html_body, 'html')
        msg.attach(html)

    # we send the message via sendmail because direct traffic to port 25
    # is firewalled off
    p = subprocess.Popen((SENDMAIL_PATH, '-t', '-oi'),
                         stdin=subprocess.PIPE)
    p.communicate(msg.as_string().encode('utf8'))


def send_problem_report(problem):
    """Send a problem report to OCF staff."""

    def format_frame(frame):
        _, filename, line, funcname, _, _ = frame
        return '{}:{} ({})'.format(filename, line, funcname)

    callstack = '\n        by '.join(map(format_frame, inspect.stack()))
    body = \
        """A problem was encountered and reported via ocflib:

{problem}

====
Hostname: {hostname}
Callstack:
    at {callstack}
""".format(problem=problem, hostname=socket.getfqdn(), callstack=callstack)

    send_mail(
        MAIL_ROOT,
        '[ocflib] Problem report from ' + socket.getfqdn(),
        body,
        sender='ocflib <root@ocf.berkeley.edu>',
    )



"""Methods for tracking meeting times across OCF.

All times are assumed to be OST (OCF Standard Time).
"""
from collections import namedtuple
from datetime import date
from datetime import datetime
from datetime import timedelta
from time import localtime

from yaml import safe_load

MEETING_HOURS_FILE = '/etc/ocf/meeting_hours.yaml'
DAYS_OF_WEEK = ['Monday', 'Tuesday', 'Wednesday',
                'Thursday', 'Friday', 'Saturday', 'Sunday']

Meeting = namedtuple(
    'Meeting',
    ['day', 'time', 'subject', 'short', 'irl', 'virtual']
)


def _load_meeting_hours():
    """Load meeting hours, from the /etc/ocf folder."""
    with open(MEETING_HOURS_FILE, 'r') as f:
        return safe_load(f)


def _get_meeting_hours():
    meeting_hours = _load_meeting_hours()

    meeting_hour_list = []
    for day in DAYS_OF_WEEK:
        if not meeting_hours['meeting-hours'][day]:
            continue

        for meeting_hour in meeting_hours['meeting-hours'][day]:
            meeting_hour_list.append(
                Meeting(
                    day=day,
                    time=_parse_hour(meeting_hour['time']),
                    subject=meeting_hour['subject'],
                    short=meeting_hour['short'],
                    irl=meeting_hour['irl'],
                    virtual=meeting_hour['virtual'],
                )
            )

    return meeting_hour_list


def _parse_hour(hour):
    """
    Converts a 2-element list of hours like ['11:00', '13:00'] to a string
    in 12-hour time, like '11:00AM - 1:00PM'.
    """
    return '{} - {}'.format(
        datetime.strptime(
            hour[0],
            '%H:%M'
        ).strftime('%-I:%M%p'),
        datetime.strptime(
            hour[1],
            '%H:%M'
        ).strftime('%-I:%M%p')
    )


def _time_to_range(hours):
    """
    Converts a time from format '11:00AM - 1:00PM' to (660, 780)
    """
    first_colon = hours.find(':')
    if first_colon != 1 and first_colon != 2:
        return

    first_half = hours[first_colon + 3:first_colon + 5]
    if first_half != 'AM' and first_half != 'PM':
        return

    hyphen = hours.find('-')
    if hyphen == -1:
        return

    second_colon = hours.find(':', first_colon + 1)
    if second_colon != hyphen + 3 and second_colon != hyphen + 4:
        return

    second_half = hours[second_colon + 3:second_colon + 5]
    if second_half != 'AM' and second_half != 'PM':
        return

    start_hours = int(hours[0:first_colon])
    start_minutes = int(hours[first_colon + 1:first_colon + 3])
    end_hours = int(hours[hyphen + 2:second_colon])
    end_minutes = int(hours[second_colon + 1:second_colon + 3])

    start_pm_offset = 12 if first_half == 'PM' else 0
    end_pm_offset = 12 if second_half == 'PM' else 0

    start_time = start_hours * 60 + start_pm_offset * 60 + start_minutes
    end_time = end_hours * 60 + end_pm_offset * 60 + end_minutes

    return (start_time, end_time)


def _iso_weekday_to_str(num):
    if num == 1:
        return 'Monday'
    elif num == 2:
        return 'Tuesday'
    elif num == 3:
        return 'Wednesday'
    elif num == 4:
        return 'Thursday'
    elif num == 5:
        return 'Friday'
    elif num == 6:
        return 'Saturday'
    elif num == 7:
        return 'Sunday'

    return ''


def _get_next_meeting(today=date.today(), now=localtime()):
    now = now.tm_hour * 60 + localtime().tm_min
    days = [(today + timedelta(days=i)).strftime('%A') for i in range(7)]

    meetings = sorted(
        _get_meeting_hours(),
        key=lambda meeting: days.index(meeting.day)
    )

    for meeting in meetings:
        ranged_time = _time_to_range(meeting.time)
        if ranged_time[0] > now or meeting.day != _iso_weekday_to_str(today.isoweekday()):
            return meeting

    return None


def _get_current_meeting(today=date.today(), now=localtime()):
    now = now.tm_hour * 60 + localtime().tm_min
    days = [(today + timedelta(days=i)).strftime('%A') for i in range(7)]

    meetings = sorted(
        _get_meeting_hours(),
        key=lambda meeting: days.index(meeting.day)
    )

    for meeting in meetings:
        ranged_time = _time_to_range(meeting.time)
        if ranged_time[0] < now and ranged_time[1] > now and meeting.day == _iso_weekday_to_str(today.isoweekday()):
            return meeting

    return None


def read_meeting_list(**kwargs):
    return _get_meeting_hours(**kwargs)


def read_next_meeting(**kwargs):
    return _get_next_meeting(**kwargs)


def read_current_meeting(**kwargs):
    return _get_current_meeting(**kwargs)



from github import Github
from github import InputGitTreeElement


class GithubCredentials():
    """Basic class to store Github credentials and verify input"""

    def __init__(self, username=None, password=None, token=None):
        if not (username or password or token):
            raise ValueError('No credentials supplied')

        if (username and password and token):
            raise ValueError('Can\'t pass in both username/password and token')

        if (username and not password) or (password and not username):
            raise ValueError('Username/password supplied but not the other')

        self.username = username
        self.password = password
        self.token = token


class GitRepo():
    """
    Extension of PyGithub with a couple of other helper methods.
    """

    def __init__(self, repo_name, credentials=None):
        """Retrieves a Repository by its fully qualified name. If credentials are passed
        they will be used."""
        if not credentials:
            self._github = Github().get_repo(repo_name)
        elif credentials.token:
            self._github = Github(credentials.token).get_repo(repo_name)
        else:
            self._github = Github(credentials.username, credentials.password).get_repo(repo_name)

    @property
    def github(self):
        """
        Direct access to the underlying PyGithub object.
        """
        return self._github

    def get_file(self, filename):
        """Fetch and decode the file from the master branch.
        Note that GitHub's API only supports files up to 1MB in size."""
        return self._github.get_contents(filename).decoded_content.decode('utf-8')

    def modify_and_branch(self, base_branch, new_branch_name, commit_message, filename, file_content):
        """Create a new branch from base_branch, makes changes to a file, and
        commits it to the new branch."""

        base_sha = self._github.get_git_ref('heads/{}'.format(base_branch)).object.sha
        base_tree = self._github.get_git_tree(base_sha)
        element = InputGitTreeElement(filename, '100644', 'blob', file_content)
        tree = self._github.create_git_tree([element], base_tree)

        parent = self._github.get_git_commit(base_sha)
        commit = self._github.create_git_commit(commit_message, tree, [parent])

        self._github.create_git_ref('refs/heads/{}'.format(new_branch_name), commit.sha)
